import onnxruntime as onnx
import os
import time
import cupy as np
import yaml
import pandas as pd
import src.SoybeanNext.utils as utils
from src.SoybeanNext.reader import Reader, one_hot


# This module is used for predicting and analysis the results.
# Onnxruntime could run on GPU or CPU, which depends on your installed version of this package.
# If you would like to run it on your GPU, uninstall 'onnxruntime' at first, then install 'onnxruntime-gpu'.


class Forward:
    def __init__(self, traits, batch_size=1):
        # the path of models, it should be like:
        # ../
        # onnx/
        #   -- model1.onnx
        #   -- modeln.onnx
        self.modelsPath = r'data/onnx/'
        self.batch_size = batch_size
        os.makedirs(self.modelsPath, exist_ok=True)
        # list of supported Quality Traits and corresponding index to value dictionary
        with open("data/p_trait.yaml") as p:
            self.data_p = yaml.safe_load(p)
        with open("data/n_trait.yaml") as n:
            self.data_n = yaml.safe_load(n)
        # traits to predict
        self.traits = list(set(traits))
        # preload needed models
        self.url = "http://xtlab.hzau.edu.cn/downloads/"
        for trait in self.traits:
            name = trait + '_best.onnx'
            # if model does not exist, download it
            if not os.path.isfile(f"{self.modelsPath + name}"):
                utils.downloads(self.modelsPath, name)
            self.models = {trait: onnx.InferenceSession(self.modelsPath + name, providers=['CUDAExecutionProvider', 'CPUExecutionProvider'])}

    # input is a list of samples
    # here, unpack the input data and split results by different trait types
    def forward(self, index_list, input_data=None):
        # transform the cupy array to numpy array (to avoid onnxruntime errors)
        input_data = np.asnumpy(input_data)
        # outputs of final results
        df = pd.DataFrame(None, columns=self.traits, index=index_list)
        # predict each trait for each sample
        for trait, model in self.models.items():
            # determine if this trait is a Quality Trait
            if trait in self.data_p.keys():
                # results
                results = []
                # indexed levels dictionary of this trait
                traitLevels = self.data_p[trait]
                # generate batched data and forward them
                for batch in self.batch_generator(input_data):
                    # raw results of this batch, its shape is batch * levels' amount
                    out = model.run(['output'], {'input': batch})[0]
                    # turn raw results to indexes
                    out = out.argmax(axis=1)
                    # merge output to list results, for matching levels in only one loop
                    results += out.tolist()
                df[trait] = [traitLevels[result] for result in results]
            # this trait is a Quantity Trait
            else:
                # results
                results = []
                traitMax = self.data_n[trait]['max']
                traitMin = self.data_n[trait]['min']
                traitDiff = traitMax - traitMin
                for batch in self.batch_generator(input_data):
                    # raw results of this batch, its shape is batch * levels' amount
                    out = model.run(['output'], {'input': batch})[0]
                    # merge output to list results, for matching levels in only one loop
                    results += out.tolist()
                # denormalize the result, use .get() method to transform cupy array to numpy array
                df[trait] = (np.array(results) * traitDiff + traitMin).get()
        return df

    # slice the input data into batches
    def batch_generator(self, input_data):
        start = 0
        # which batch is running now
        times = 1
        while start < input_data.shape[0]:
            end = start + self.batch_size
            # load input data batch by batch
            yield input_data[start:end, :, :, :]
            start = end
            times += 1

    # output DataFrame to .csv
    def output_dataframe(self, dataframe):
        dataframe.to_csv("./result.csv")

    # the pipeline of whole execution, return result DataFrame and flag isMissing
    def run(self, df_path, output=True):
        # instantiate class Reader
        r = Reader()
        s = time.time()
        # get the processed dataframe
        df = r.readVCF(rf"{df_path}")
        e = time.time()
        print(f"readVCF————{e - s:.2f}s")
        s = time.time()
        # get the filtered dataframe
        df_filter, isMissing = r.SNPfilter(df)
        e = time.time()
        print(f"SNPfilter————{e - s:.2f}s")
        s = time.time()
        # convert to one-hot matrix and resize every samples
        index_list = r.indexes
        sample_resized = one_hot(df_filter.values)
        e = time.time()
        print(f"one_hot————{e - s:.2f}s")
        s = time.time()
        # predict and get results
        df = self.forward(index_list, sample_resized)
        e = time.time()
        print(f"forward————{e - s: .2f}s")
        if output:
            self.output_dataframe(df)
        return df, isMissing



import warnings
import torch
from src.SoybeanNext.remodel.weight_map import weight_decoder
from src.SoybeanNext.remodel.remodel import remodel
from src.SoybeanNext.utils import outpath
from torch.utils.data import DataLoader
from sklearn import metrics
from data_process import *
from eval import *


def init_weights(net):
    if type(net) == torch.nn.Linear or type(net) == torch.nn.Conv2d:
        torch.nn.init.xavier_uniform_(net.weight)


class data_loader(torch.utils.data.Dataset):
    def __init__(self, data, label):
        self.data = torch.from_numpy(data)
        self.label = torch.from_numpy(label)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        genotype = self.data[index].float()
        label = self.label[index].float()
        return genotype, label


# In this module, users could train their own datasets on our baseline
class Train:
    def __init__(self, vcf_path, trait_path, percentage_of_train=0.7, num_workers=8, batch_size=20):
        # ignore warnings
        warnings.filterwarnings("ignore")
        # select which device to train
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        # percentage of train dataset
        self.percentage_of_train = percentage_of_train
        # the training batch size
        self.batch_size = batch_size
        # workers amount
        self.num_workers = num_workers
        # output path
        self.saved_path = outpath('train')
        # read .vcf file and .csv trait file, and instantiate the data process class
        self.d = DataProcess(vcf_path, trait_path)
        # process and return the Quality Traits dictionary and Quantity Traits dictionary
        self.p_trait_dic, self.n_trait_dic = self.d.convert_trait(f"{self.saved_path}/configs/")
        print(f"yaml configs are saved in {self.saved_path}/configs/")
        # turn keys to lists
        self.p_trait_list, self.n_trait_list = list(self.p_trait_dic.keys()), list(self.n_trait_dic.keys())



    # create train and test dataloader for training
    def dataloader(self, trait, is_quality):
        # split dataset by percentage of dataset
        train_data, train_label, test_data, test_label = self.d.to_dataset(trait, percentage=self.percentage_of_train, is_quality=is_quality)
        # dataloader of train dataset
        train_dataloader = DataLoader(data_loader(train_data, train_label), batch_size=self.batch_size, shuffle=True, num_workers=self.num_workers)
        # dataloader of test dataset
        test_dataloader = DataLoader(data_loader(test_data, test_label), batch_size=self.batch_size, shuffle=True, num_workers=self.num_workers)
        return train_dataloader, test_dataloader

    # train an epoch
    def train_for_epoch(self, train_dataloader, updater, loss, net):
        loss_ = 0.0
        for num_data, (genomap, target_trait) in enumerate(train_dataloader):
            genomap, target_trait = genomap.to(self.device), target_trait.to(self.device)
            trait_hat = net(genomap)
            loss_for_batch = loss(trait_hat, target_trait.long())
            loss_ += loss_for_batch
            if isinstance(updater, torch.optim.Optimizer):
                updater.zero_grad()
                loss_for_batch.backward()
                updater.step()
        return loss_ / (num_data + 1)

    # train the Quantity Traits
    def train_n(self, percentage=0.05, epoch=5, weight_decay=1e-5, draw=True):
        epoch_total = epoch
        weight_save_path = f"{self.saved_path}/weight/"
        pic_save_path = f"{self.saved_path}/"
        os.makedirs(weight_save_path, exist_ok=True)
        net, decoder = remodel("../data/model.yaml", 1)
        net.to(self.device)
        # initialize weights
        net.apply(init_weights)
        # set Smooth S1 Loss as the loss function
        loss = torch.nn.SmoothL1Loss()
        # set updater with Adam algorithm
        updater = torch.optim.Adam(net.parameters(), weight_decay=weight_decay)
        for trait in self.n_trait_list:
            # create the evaluation dictionary, Quantity Traits and Quality Traits have different evaluation indexes
            eval_dict = {'train_loss': [], 'test_loss': [], 'mse': [], 'r': [], }
            # initialize the loss function
            mse_loss = torch.nn.MSELoss()
            # train and test dataloaders
            train_dataloader, test_dataloader = self.dataloader(trait, is_quality=False)
            # train each epoch
            # the minimum coefficient value
            min_r = 0.0
            while epoch:
                print(f"Now training epoch {epoch_total - epoch + 1}")
                # set the model as train mode
                net.train()
                # train this epoch and compute the average loss
                avg_train_loss = self.train_for_epoch(train_dataloader, updater, loss, net)
                # set the model as evaluation mode
                net.eval()
                # for the test dataset
                with torch.no_grad():
                    # initialize the evaluation indexes
                    mse = 0.0
                    loss_test = 0.0
                    hat = np.array([])
                    truth = np.array([])
                    for index, (test_genomap, test_trait) in enumerate(test_dataloader):
                        # load the feature maps and traits on device
                        test_genomap, test_trait = test_genomap.to(self.device), test_trait.to(self.device)
                        # the predict results
                        y_hat = net(test_genomap)
                        # compute loss on test dataset
                        loss_test += loss(y_hat, test_trait.long())
                        # reshape y_hat as an 1-dimension vector, and insert it to hat
                        hat = np.insert(hat, 0, y_hat.to('cpu').detach().numpy().reshape(len(y_hat), ), axis=0)
                        # reshape acc as an 1-dimension vector, and insert it to acc
                        truth = np.insert(truth, 0, test_trait.to('cpu').numpy(), axis=0)
                        # compute MSE between predict results and actual values
                        mse += mse_loss(y_hat, test_trait)
                # count evaluation indexes
                loss_test = loss_test / (index + 1)
                eval_dict['test_loss'].append(float(loss_test.to('cpu').detach().numpy()))
                eval_dict['train_loss'].append(float(avg_train_loss.to('cpu').detach().numpy()))
                # the average MSE
                mse / (index + 1)
                # compute the correlation coefficient value
                r = np.corrcoef(hat, truth)[0][1]
                # append the coefficient value into evaluation dictionary
                eval_dict['r'].append(r)
                # save the best weight with the highest coefficient value
                if r >= min_r:
                    torch.save(net, os.path.join(weight_save_path, f'{trait}_best.pt'))
                    # refresh the minimum coefficient value
                    min_r = eval_dict['r'][-1]
                epoch -= 1
            print(f"weight are saved in {weight_save_path}/{trait}_best.pt")
            # set the trait of this evaluation dictionary
            eval_dict['trait'] = trait
            if draw:
                Eval(eval_dict, pic_save_path)
                print(f"pictures are saved in {pic_save_path}")
        snp_pos = weight_decoder(net,decoder,self.d.columns, percentage=percentage)
        print(f"Top {percentage*100}% most important SNP positions are: {snp_pos}")
        return eval_dict

    # train the Quality Traits
    def train_p(self, percentage=0.05, epoch=5, weight_decay=1e-5, draw=True):
        epoch_total = epoch
        weight_save_path = f"{self.saved_path}/weight/"
        pic_save_path = f"{self.saved_path}/"
        os.makedirs(weight_save_path, exist_ok=True)
        for trait in self.p_trait_list:
            # train and test dataloaders
            train_dataloader, test_dataloader = self.dataloader(trait, is_quality=True)
            # get how many levels this trait has
            num_classes = len(self.p_trait_dic[trait])
            # load the model
            # for different class amounts, the net has to be reloaded in each loop
            net, decoder = remodel("../data/model.yaml", num_classes)
            net.to(self.device)
            # initialize weights
            net.apply(init_weights)
            # set Entropy Loss as the loss function
            loss = torch.nn.CrossEntropyLoss()
            # set updater with Adam algorithm
            updater = torch.optim.Adam(net.parameters(), weight_decay=weight_decay)
            # evaluation dictionary of Quality Traits
            eval_dict = {'train_loss': [], 'test_loss': [], 'acc': [], 'recall': [], 'precision': [], 'f1_score': []}
            # the minimum evaluation value
            eval_value = 0.0
            # train each epoch
            while epoch:
                print(f"Now training epoch {epoch_total - epoch + 1}")
                # set the model as train mode
                net.train()
                # train this epoch and compute the average loss
                avg_train_loss = self.train_for_epoch(train_dataloader, updater, loss, net)
                # set the model as evaluation mode
                net.eval()
                # for the test dataset
                with torch.no_grad():
                    # initialize the evaluation indexes
                    acc_score = 0.0
                    recall_score = 0.0
                    f1_score = 0.0
                    precision = 0.0
                    loss_test = 0.0
                    hat = np.array([])
                    truth = np.array([])
                    for index, (test_genomap, test_trait) in enumerate(test_dataloader):
                        # load the feature maps and traits on device
                        test_genomap, test_trait = test_genomap.to(self.device), test_trait.to(self.device)
                        # the predict results
                        y_hat = net(test_genomap)
                        # compute loss on test dataset
                        loss_test += loss(y_hat, test_trait.long())
                        # reshape y_hat as an 1-dimension vector, and insert it to hat
                        hat = np.insert(hat, 0, np.argmax(y_hat.to('cpu').detach().numpy(), axis=1), axis=0)
                        # reshape acc as an 1-dimension vector, and insert it to acc
                        truth = np.insert(truth, 0, test_trait.to('cpu').numpy(), axis=0)
                        # compute accuracy scores
                        acc_score += metrics.accuracy_score(truth, hat)
                        # compute recall scores
                        recall_score += metrics.recall_score(truth, hat, average='macro')
                        # compute F1 scores
                        f1_score += metrics.f1_score(truth, hat, average='macro')
                        # compute precision scores
                        precision += metrics.precision_score(truth, hat, average='macro')
                # count evaluation indexes
                loss_test = loss_test / (index + 1)
                eval_dict['test_loss'].append(float(loss_test.to('cpu').detach().numpy()))
                eval_dict['train_loss'].append(float(avg_train_loss.to('cpu').detach().numpy()))
                eval_dict['acc'].append(acc_score / (index + 1))
                eval_dict['recall'].append(recall_score / (index + 1))
                eval_dict['f1_score'].append(f1_score / (index + 1))
                eval_dict['precision'].append(precision / (index + 1))
                # save the best weight with the highest coefficient value
                value = np.array(list(eval_dict.values()))[2:, -1].mean()
                if value > eval_value:
                    # compute the confuse matrix
                    truth = list(map(int, truth))
                    hat = list(map(int, hat))
                    confusion_matrix = metrics.confusion_matrix(truth, hat, labels=list(self.p_trait_dic[trait].keys()), sample_weight=None)
                    # refresh the evaluation value
                    eval_value = value
                    torch.save(net, os.path.join(weight_save_path, f'{trait}_best.pt'))
                epoch -= 1
            print(f"weight are saved in {weight_save_path}/{trait}_best.pt")
            # set the trait of this evaluation dictionary and save the confusion matrix
            eval_dict['trait'] = trait
            eval_dict['confusion_matrix'] = confusion_matrix
            if draw:
                Eval(eval_dict, pic_save_path, self.p_trait_dic)
                print(f"pictures are saved in {pic_save_path}")
        snp_pos = weight_decoder(net,decoder,self.d.columns, percentage=percentage)
        print(f"Top {percentage * 100}% most important SNP positions are: {snp_pos}")
        return eval_dict


# an example
# t = Train("../data/test.vcf", "../data/final_trait.csv")
# t.train_p()



import torch
from torch import nn


class CA_attention(nn.Module):
    def __init__(self, channel, h, w, reduction=16):
        super(CA_attention, self).__init__()

        self.h = h
        self.w = w

        self.avg_pool_x = nn.AdaptiveAvgPool2d((h, 1))
        self.avg_pool_y = nn.AdaptiveAvgPool2d((1, w))

        self.conv_1x1 = nn.Conv2d(in_channels=channel, out_channels=channel // reduction, kernel_size=1, stride=1,
                                  bias=False)

        self.relu = nn.ReLU()
        self.bn = nn.BatchNorm2d(channel // reduction)

        self.F_h = nn.Conv2d(in_channels=channel // reduction, out_channels=channel, kernel_size=1, stride=1,
                             bias=False)
        self.F_w = nn.Conv2d(in_channels=channel // reduction, out_channels=channel, kernel_size=1, stride=1,
                             bias=False)

        self.sigmoid_h = nn.Sigmoid()
        self.sigmoid_w = nn.Sigmoid()

    def forward(self, x):
        x_h = self.avg_pool_x(x).permute(0, 1, 3, 2)
        x_w = self.avg_pool_y(x)

        x_cat_conv_relu = self.relu(self.conv_1x1(torch.cat((x_h, x_w), 3)))

        x_cat_conv_split_h, x_cat_conv_split_w = x_cat_conv_relu.split([self.h, self.w], 3)

        s_h = self.sigmoid_h(self.F_h(x_cat_conv_split_h.permute(0, 1, 3, 2)))
        s_w = self.sigmoid_w(self.F_w(x_cat_conv_split_w))

        out = x * s_h.expand_as(x) * s_w.expand_as(x)

        return out


class ChannelAttention(nn.Module):
    def __init__(self, in_channels, reduction_ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(in_channels, in_channels // reduction_ratio),
            nn.ReLU(inplace=True),
            nn.Linear(in_channels // reduction_ratio, in_channels)
        )
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * self.sigmoid(y)


class SpatialAttention(nn.Module):
    def __init__(self):
        super(SpatialAttention, self).__init__()
        self.conv = nn.Conv2d(2, 1, kernel_size=7, padding=3)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        max_pool = torch.max(x, dim=1, keepdim=True)[0]
        avg_pool = torch.mean(x, dim=1, keepdim=True)
        y = torch.cat([max_pool, avg_pool], dim=1)
        y = self.conv(y)
        return x * self.sigmoid(y)


class CBAM_attention(nn.Module):
    def __init__(self, in_channels, reduction=16):
        super(CBAM_attention, self).__init__()
        self.channel_attention = ChannelAttention(in_channels, reduction)
        self.spatial_attention = SpatialAttention()

    def forward(self, x):
        x = self.channel_attention(x)
        x = self.spatial_attention(x)
        return x


class SE_attention(nn.Module):
    def __init__(self, channel, reduction=16):
        super(SE_attention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, channel // reduction),
            nn.ReLU(inplace=True),
            nn.Linear(channel // reduction, channel),
            nn.Sigmoid()
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


class CNN_Block(nn.Module):
    def __init__(self, in_channel, out_channel, kernel_size, padding, stride, drop_out=0.3):
        super().__init__()
        if padding != 0:
            padding_mode = 'reflect'
            self.block = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, kernel_size, padding=padding, padding_mode=padding_mode,
                          stride=stride, bias=False),
                nn.BatchNorm2d(out_channel),
                nn.Dropout(drop_out),
            )
        else:
            self.block = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, kernel_size, stride=stride, bias=False),
                nn.BatchNorm2d(out_channel),
                nn.Dropout(drop_out),
            )

    def forward(self, x):
        return self.block(x)


class Linear_(nn.Module):
    def __init__(self, length, out_features, drop_out=0.3):
        super().__init__()
        self.linear = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(drop_out),
            nn.ReLU(),
            nn.Linear(length, out_features, bias=False)
        )

    def forward(self, x):
        return self.linear(x)


class ReLU_(nn.Module):
    def __init__(self):
        super().__init__()
        self.relu = nn.Sequential(
            nn.ReLU(),
        )

    def forward(self, x):
        return self.relu(x)


class Rediual_Block(nn.Module):
    def __init__(self, in_channel, out_channel, kernel_size, padding, stride, drop_out):
        super().__init__()
        self.stride = stride
        self.c = nn.Sequential(
            nn.Conv2d(in_channel, out_channel, kernel_size, padding=padding, padding_mode='reflect', stride=self.stride,
                      bias=False),
            nn.BatchNorm2d(out_channel),
            nn.Dropout(drop_out),
            nn.Conv2d(out_channel, out_channel, 3, 1, padding_mode='reflect', stride=1, bias=False),
            nn.BatchNorm2d(out_channel),
            nn.Dropout(drop_out),

        )
        if self.stride > 1:
            if padding != 0:
                self.r = nn.Sequential(
                    nn.Conv2d(in_channel, out_channel, kernel_size, padding=padding, padding_mode='reflect',
                              stride=self.stride, bias=False),
                    nn.BatchNorm2d(out_channel)
                )
            else:
                self.r = nn.Sequential(
                    nn.Conv2d(in_channel, out_channel, kernel_size, stride=stride, bias=False),
                    nn.BatchNorm2d(out_channel)
                )

    def forward(self, x):
        if self.stride > 1:
            o1 = self.c(x)
            out = self.r(x) + o1
        else:
            out = self.c(x)
        return out


class TransCNN_Block(nn.Module):
    def __init__(self, in_channel, out_channel, kernel_size, padding, stride):
        super().__init__()
        if padding != 0:
            padding_mode = 'reflect'
            self.block = nn.Sequential(
                nn.ConvTranspose2d(in_channel, out_channel, kernel_size, padding=padding, padding_mode=padding_mode,
                                   stride=stride, bias=False),
            )
        else:
            self.block = nn.Sequential(
                nn.ConvTranspose2d(in_channel, out_channel, kernel_size, stride=stride, bias=False),
            )

    def forward(self, x):
        return self.block(x)


class Reshape(nn.Module):
    def __init__(self, *args):
        super(Reshape, self).__init__()
        self.shape = args

    def forward(self, x):
        return x.view((x.size(0),) + self.shape)



import matplotlib.pyplot as plt
import numpy as np
import yaml
import os

from datetime import datetime


# In this module, the evaluation dictionary after running train method will be drawn.
class Eval:
    def __init__(self, eval_dict, save_dir, p_trait_dic=None):
        self.eval_dict = eval_dict
        self.save_dir = save_dir
        self.p_trait_dic = p_trait_dic
        self.model_eval()
        self.pic_draw()

    def model_eval(self):
        dict_ = {}
        trait = self.eval_dict['trait']
        if not self.p_trait_dic:
            best_model_index = self.eval_dict['r'].index(max(self.eval_dict['r']))
            for key in self.eval_dict.keys():
                dict_[key] = self.eval_dict[key][best_model_index]
            with open(f"{self.save_dir}configs/{trait}_module_eval.yaml", 'w') as f:
                f.write(yaml.dump(dict_, allow_unicode=True))
        else:
            temp_dict = dict([(key, self.eval_dict[key]) for key in ['acc', 'precision', 'recall', 'f1_score']])
            eval_data = np.array(list(temp_dict.values()))
            eval_data = np.sum(eval_data, axis=0)
            best_model_index = np.argmax(eval_data)
            for key in self.eval_dict.keys():
                if key != 'confusion_matrix' and key != 'trait':
                    dict_[key] = self.eval_dict[key][best_model_index]
                else:
                    dict_[key] = self.eval_dict[key]
            with open(f"{self.save_dir}configs/{trait}_module_eval.yaml", 'w') as f:
                f.write(yaml.dump(dict_, allow_unicode=True))

    def pic_draw(self):
        x = [i for i in range(1, len(self.eval_dict['train_loss']) + 1)]
        train_loss = self.eval_dict['train_loss']
        test_loss = self.eval_dict['test_loss']
        trait = self.eval_dict['trait']
        if not self.p_trait_dic:
            mse = self.eval_dict['mse']
            r = self.eval_dict['r']

            # loss 曲线
            plt.plot(x, train_loss, label='train', linewidth=0.9)
            plt.plot(x, test_loss, label='test', linewidth=0.9)
            plt.xlabel('Epoch', fontsize=13)
            plt.ylabel('SmoothL1 Loss', fontsize=13)
            plt.legend(loc='best')
            plt.savefig(os.path.join(self.save_dir, f'{trait}_loss.svg'))
            plt.close()
            # mse曲线
            plt.plot(x, mse, linewidth=0.9)
            plt.xlabel('Epoch', fontsize=13)
            plt.ylabel('Meansquare Error', fontsize=13)
            plt.savefig(os.path.join(self.save_dir, f'{trait}_mse.svg'))
            plt.close()
            # 相关系数曲线
            plt.plot(x, r, linewidth=0.9)
            plt.xlabel('Epoch', fontsize=13)
            plt.ylabel('Pearson Correlation Coefficient', fontsize=13)
            plt.savefig(os.path.join(self.save_dir, f'{trait}_r.svg'))
            plt.close()

        else:
            acc = self.eval_dict['acc']
            pre = self.eval_dict['precision']
            recall = self.eval_dict['recall']
            f1_score = self.eval_dict['f1_score']
            confusion_matrix = self.eval_dict['confusion_matrix']

            # loss曲线
            plt.plot(x, train_loss, label='train', linewidth=0.9)
            plt.plot(x, test_loss, label='test', linewidth=0.9)
            plt.xlabel('Epoch', fontsize=13)
            plt.ylabel('CrossEntrpy Loss', fontsize=13)
            plt.legend(loc='best')
            plt.savefig(os.path.join(self.save_dir, f'{trait}_loss.svg'))
            plt.close()
            # eval曲线
            plt.plot(x, acc, label='accuracy', linewidth=0.9)
            plt.plot(x, pre, label='precision', linewidth=0.9)
            plt.plot(x, f1_score, label='f1_score', linewidth=0.9)
            plt.plot(x, recall, label='recall', linewidth=0.9)
            plt.xlabel('Epoch', fontsize=13)
            plt.ylabel('Values', fontsize=13)
            plt.legend(loc='best')
            plt.savefig(os.path.join(self.save_dir, f'{trait}_eval.svg'))
            plt.close()

            print(confusion_matrix)
            # 混淆矩阵
            confusion_matrix = confusion_matrix.astype('float') / confusion_matrix.sum(axis=1)[:, np.newaxis]  # 归一化
            print(confusion_matrix)
            x = list(self.p_trait_dic[trait].keys())
            plt.imshow(confusion_matrix, interpolation='nearest', cmap=plt.get_cmap('Blues'))
            plt.xticks(x, list(self.p_trait_dic[trait].values()), rotation=45)  # 将标签印在x轴坐标上， 并倾斜45度
            plt.yticks(x, list(self.p_trait_dic[trait].values()))  # 将标签印在y轴坐标上
            plt.ylabel('True label', fontsize=13)
            plt.xlabel('Predicted label', fontsize=13)
            for i in range(np.shape(confusion_matrix)[0]):
                for j in range(np.shape(confusion_matrix)[1]):
                    if int(confusion_matrix[i][j] * 100 + 0.5) > 0:
                        plt.text(j, i, format(int(confusion_matrix[i][j] * 100 + 0.5), 'd') + '%',
                                 ha="center", va="center",
                                 color="white" if confusion_matrix[i][j] > 0.7 else "black")
            plt.savefig(os.path.join(self.save_dir, f'{trait}_confusion_matrix.svg'))
            plt.close()



import pandas as pd
import numpy as np

# This module is a CPU implement of reader.py

class Reader_CPU:
    # initial class Reader, load the needed SNP file
    def __init__(self, SNPlistPath="../data/snp.txt"):
        # open the file and transform it to list
        with open(SNPlistPath) as f:
            # delete '\n' in every elements
            self.SNPlist = [snp.rstrip() for snp in f.readlines()]
        self.columns = []
        self.indexes = None

    # the CPU implement of readVCF method
    def readVCF(self, vcf_path, reset=False):
        # open the .vcf file
        with open(vcf_path, "r") as f:
            # if the line does not start as "##", this line is the header line
            header = next((i for i, line in enumerate(f) if not line.startswith('##')), 0)
        # read the file from header row
        df = pd.read_csv(vcf_path, header=header, sep='\t', dtype=str)
        # clean the useless columns, such as ['QUAL', 'FILTER', 'INFO', 'FORMAT', '#CHROM', 'POS', 'REF', 'ALT']
        # after doing that, only ID and sample IDs will be kept
        self.columns.append(df.columns[2])
        self.columns.extend(df.columns[9:])
        # concat the chromosome name and position as new ID for SNPs
        df['ID'] = df['#CHROM'].str.cat(df['POS'], sep='_')
        # set ID column as index and drop the useless columns
        df = df[self.columns].set_index('ID')
        # clear values of DataFrame
        # sometimes, values might be with annotations, for example, '1/1(some annotation)', which should be deleted
        for column in self.columns[1:]:
            # only the first three characters could be kept
            df[column] = df[column].str.slice(0, 3)
            # create a temporary series, which is replaced by rules
            temp_series = df[column].replace({r'1/1': '1', r'1|1': '1', r'0/1': '2', r'0|1': '2'})
            # make a mask to find out unchanged elements
            mask = df[column] == temp_series
            # assign '3' for unchanged elements
            temp_series[mask] = '3'
            # replace column of DataFrame with assigned series
            df[column] = temp_series.astype(np.int32)
            # delete useless variables
            del temp_series, mask
        # transpose the DataFrame, now the SNP IDs are columns, and sample IDs are indexes
        df = df.transpose()
        # keep the indexes and columns
        self.indexes = df.index
        self.columns = df.columns
        # drop indexes and reduce memory consumption
        if reset:
            df.reset_index(drop=True, inplace=True)
        return df

    # the CPU implement of SNPfilter method
    def SNPfilter(self, dataframe):
        # find out the missing SNP positions
        missing = list(set(self.SNPlist).difference(set(self.columns)))
        # set a flag to determine if missing is an empty set
        isMissing = bool(missing)
        # determine if there are missing values
        if isMissing:
            # fill the missing values with './.'
            df_missing = pd.DataFrame(None, columns=missing, index=self.indexes, dtype=np.int32)
            df_missing[:] = 3
            # concat the original dataframe and filling dataframe
            df_filter = pd.concat([dataframe, df_missing], axis=1)
            # delete useless variable
            del df_missing

        else:
            # if there are not missing values, straightly reset the columns
            df_filter = dataframe[self.SNPlist].astype(np.int32)
        # delete useless variable
        del dataframe, missing
        return df_filter, isMissing

# the CPU implement of one_hot method
def one_hot(matrix):
    # repeat 3 times to 3D, in this way, transforming to one-hot matrix does not need explicit loop, which could perform better
    matrix_3D = np.zeros((matrix.shape[0], matrix.shape[1], 3))
    # the rule for one-hot is like:
    #             '1/1' or '1|1'        '0/1' or '0|1'        other
    # channel 1          1                     1                0
    # channel 2          1                     0                1
    # channel 3          0                     1                1
    # for the first layer, only the first and second condition will be valued as 1
    matrix_3D[:, :, 0] = np.where(np.isin(matrix[:, :], np.array([1, 2])), 1, 0)
    # for the second layer, only the second condition will be valued as 0
    matrix_3D[:, :, 1] = np.where(np.isin(matrix[:, :], np.array(2)), 0, 1)
    # for the third layer, only the first condition will be valued as 0
    matrix_3D[:, :, 2] = np.where(np.isin(matrix[:, :], np.array(1)), 0, 1)
    # resize every sample to shape of model input
    # set an index
    n = -1
    # create batches of resized matrix
    sample_resized = np.empty((matrix.shape[0], 3, 206, 206)).astype(np.float32)
    for sample in matrix_3D:
        # !!!!!(cite from numpy v1.25 document)
        # If the new array is larger than the original array, then the new array is filled with repeated copies of a.
        # Note that this behavior is different from a.resize(new_shape) which fills with zeros instead of repeated copies of a.
        # !!!!!
        sample = np.resize(sample, (206, 206, 3))
        # transpose the order of sample's channels, for cp.reshape might change its order
        # for example, before reshape, sample has order [a, b, c], which might be changed after reshape, like [1, c, b, a]
        # to keep the result's order, here a transposition is significant
        sample = np.transpose(sample, (2, 0, 1))
        # the first dimension is batch size, just for fitting the input shape
        sample = np.reshape(sample, (1, 3, 206, 206))
        n += 1
        sample_resized[n, :, :, :] = sample
    return sample_resized



import numpy as np
import torch
from torch.nn.parameter import Parameter


def weight_decoder(net, decoder, snp_list, percentage):
    snp = snp_list
    net_weight_list = []
    # 权重赋值
    for layer in net.modules():
        if isinstance(layer, torch.nn.Conv2d):
            net_weight_list.append(layer.weight)
        elif isinstance(layer, torch.nn.Linear):
            net_weight_list.append(Parameter(layer.weight.T))
    net_weight_list = net_weight_list[::-1]
    num_class = 1 if np.array(net_weight_list[0].shape).min() == np.array(net_weight_list[0].shape).max() else np.array(
        net_weight_list[0].shape).min()
    decoder.to('cuda')
    decoder.eval()

    # 特征解码
    with torch.no_grad():
        i = 0
        for layer in decoder.modules():
            if isinstance(layer, torch.nn.ConvTranspose2d) or isinstance(layer, torch.nn.Linear):
                layer.weight = net_weight_list[i]
                i += 1
        x = (torch.zeros((1, num_class)) + 1).to('cuda')
        y = decoder(x)
        y = torch.sum(y, dim=1).to('cpu').numpy()
        data = y.reshape((206 * 206))[:len(snp) + 1]
        # 获取位点
        snp_pos = np.array(np.where(data >= np.percentile(data, 1 - percentage)))
        snp_pos = [snp[pos] for pos in snp_pos[0]]
        return snp_pos



import pandas as pd
import yaml
import os
from src.SoybeanNext.reader_cpu import Reader_CPU, one_hot
from sklearn.model_selection import train_test_split
from datetime import datetime


# In this module, .csv file would be read and split into test and train datasets.
# A .yaml config file should be provided. Due to randomly split, datasets might be different.
class DataProcess:

    def __init__(self, genotype_file, trait_file):
        # path of the .vcf genotype file
        self.genotype_path = genotype_file
        # path of the .csv traits file
        self.trait_path = trait_file
        # the trait file DataFrame, be assigned after running method convert_trait
        self.trait = None
        # the SNP IDs
        self.columns = []

    # extract trait value
    def convert_trait(self, save_path):
        # the .csv file saves traits and data, for example:
        # acid SCN3 protein
        #  id1   1      1
        #  idn   n      n
        # load the csv file
        df = pd.read_csv(self.trait_path, encoding=u'gbk')
        # load the traits config file, for example
        # p:
        #  protein
        # n:
        #  SN3
        with open('../data/traits.yaml') as t:
            traits = yaml.safe_load(t)
        # p_trait_dict: {'trait1': {0:level a, 1:level b...}...}
        # n_trait_dict: {'trait1': {max:value1, min:value2}...}
        p_trait_dict = {}
        n_trait_dict = {}
        # the Quality Traits
        if traits['p']:
            quality = df[traits['p']]
            # keep different levels of this trait
            for column in quality.columns:
                # avoid empty value to be added as a level
                tmp = df[column].dropna().drop_duplicates()
                trait_dict = tmp.reset_index(drop=True).to_dict()
                swapped_dict = {value: key for key, value in trait_dict.items()}
                df[column] = df[column].replace(swapped_dict)
                p_trait_dict[column] = trait_dict
            del tmp

        # the Quantity Traits
        if traits['n']:
            quantity = df[traits['n']]
            n_min = quantity.min()
            n_max = quantity.max()
            # normalize the input data for training
            df[traits['n']] = (quantity - n_min) / (n_max - n_min)
            # keep the information of max value and minimum value of this trait for denormalization
            for trait in n_min.index:
                n_trait_dict[trait] = {"max": float(n_max[trait]), "min": float(n_min[trait])}
        # save dictionaries as .yaml files
        os.makedirs(save_path, exist_ok=True)
        with open(f'{save_path}/n_trait.yaml', 'w') as n_file:
            n_file.write(yaml.dump(n_trait_dict))
        with open(f'{save_path}/p_trait.yaml', 'w') as p_file:
            p_file.write(yaml.dump(p_trait_dict))
        self.trait = df.set_index('acid')
        return p_trait_dict, n_trait_dict

    # split dataset into train and test dataset, and encode genotype data by one-hot method
    def to_dataset(self, trait_to_train, percentage, is_quality):
        # bool list of non-missing values
        mask = self.trait[trait_to_train].notna()
        # IDs of non-missing values
        IDs = self.trait.index[mask].tolist()
        # ignore missing values, and transform to numpy array
        trait_value = self.trait[trait_to_train][mask].values
        r = Reader_CPU()
        # read the .vcf genotype file
        genotype_original = r.readVCF(self.genotype_path)
        # the SNP IDs
        self.columns = genotype_original.columns.to_list()
        # ignore missing values, and transform to numpy array
        genotype_value = genotype_original.loc[IDs, :].values
        # if trait_to_train is a quality trait, then split each level by the same percentage
        # finally, make genotype data one-hot encoded
        if is_quality:
            train_data, test_data, train_label, test_label = train_test_split(genotype_value, trait_value, train_size=percentage, stratify=trait_value)
            train_data = one_hot(train_data)
            test_data = one_hot(test_data)
        else:
            train_data, test_data, train_label, test_label = train_test_split(genotype_value, trait_value, train_size=percentage)
            train_data = one_hot(train_data)
            test_data = one_hot(test_data)
        return train_data, train_label, test_data, test_label

# an example
# d = DataProcess("../data/test.vcf", "../data/final_trait.csv")
# d.convert_trait()




import yaml
from src.SoybeanNext.remodel.SoyDNGP import *


def remodel(path, num_classes, show_structure=True):
    # 读取yaml
    with open(path, 'r') as f:
        dic = yaml.load(f, Loader=yaml.FullLoader)
    layer_dic = dic['model']
    layer_list = []
    # 根据yaml字典实例化net
    for key, value in layer_dic.items():
        layer_list.append(eval(key.split('.')[0] + f'{value}'))
    net = nn.Sequential(*layer_list)
    # 根据实例化的net生成对应decoder 注意仅对卷积和全连接层有效
    layer_list = []
    for layer in net.modules():
        if isinstance(layer, torch.nn.Conv2d):
            layer_list.append(nn.ConvTranspose2d(in_channels=layer.out_channels,
                                                 out_channels=layer.in_channels,
                                                 kernel_size=layer.kernel_size,
                                                 padding=layer.padding,
                                                 stride=layer.stride))
        elif isinstance(layer, torch.nn.Linear):
            out_channel = layer_list[-1].in_channels
            out_length = layer.in_features
            layer_list.append(
                Reshape(out_channel, int((out_length / out_channel) ** 0.5), int((out_length / out_channel) ** 0.5)))
            layer_list.append(nn.Linear(in_features=layer.out_features, out_features=layer.in_features))
    if show_structure:
        print(f"Your model structure is: \n{net}")
    layer_list = layer_list[::-1]
    decoder = nn.Sequential(*layer_list)
    return net, decoder



import os
from tqdm import tqdm
import requests
import torch


# The utils module contains kinds of tools.

# In this module, the pointed file could be downloaded.
# more examples would be added in future
# the parameter path means the root path of downloaded files, for example, if you want to download to path 'onnx':
# --onnx
#   --model1.onnx
#   --model2.onnx
# you should assign it as 'onnx/'
def downloads(path, name):
    url = "http://xtlab.hzau.edu.cn/downloads/"
    # create the folder from parameter path, if this folder has already existed, do nothing
    os.makedirs(path, exist_ok=True)
    # the total size of this model
    response = requests.get(url + name)
    total_size = int(response.headers.get('content-length', 0))
    # download with progress bar by tqdm and requests
    with open(path + name, 'wb') as file, tqdm(
            desc=name,
            total=total_size,
            unit='iB',
            unit_scale=True,
            unit_divisor=1024,
    ) as bar:
        for data in response.iter_content(chunk_size=1024):
            size = file.write(data)
            bar.update(size)

# This module is used for transforming .pt weights to .onnx weights for better performance.
# The inputPath should be like:
# ../
# inputPath/
#   -- weight1.pt
#   -- weightn.pt
# Results will be outputted to inputPath/../onnx

# transform .pt weights to .onnx weights, input_channel, input_W and input_H should be given
# set cuda switch True to load model and dummy tensor on cuda device, using cpu otherwise
def exportAsONNX(inputPath, input_W=206, input_H=206, input_channel=3, cuda=True):
    # path of the weights
    inputPath = inputPath
    # loop through files in the path
    for name in os.listdir(inputPath):
        # determine whether it is .pt weight
        if ".pt" in name:
            # print its name
            print(name)
            # load the weight
            model = torch.load(f"{inputPath}/{name}", map_location=torch.device('cuda' if cuda else 'cpu'))
            # generate a random input tensor
            dummy_input = torch.randn(1, input_channel, input_W, input_H).to('cuda' if cuda else 'cpu')
            # determine whether the output path exists
            if not os.path.exists(f"{inputPath}/../onnx"):
                # if not exists, create the folder
                os.mkdir(f"{inputPath}/../onnx")
            # export the model to .onnx, the first size is dynamic, which means it supports batch-input
            dynamic_axes = {'input': {0: 'batch_size'}, 'output': {0: 'batch_size'}}
            torch.onnx.export(model, dummy_input, f"{inputPath}/../onnx/{name.replace('pt', 'onnx')}", input_names=['input'], output_names=['output'], dynamic_axes=dynamic_axes)


# set a unified path name
def outpath(work):
    os.makedirs(f'../runs/{work}/', exist_ok=True)
    folder_names = os.listdir(f'../runs/{work}/')
    if folder_names:
        number_list = [int(folder_name[len(work):]) for folder_name in folder_names if folder_name.startswith(work)]
        number = max(number_list) + 1
    else:
        number = 1
    new_path = f'../runs/{work}/{work}{number}'
    os.makedirs(new_path)
    print(f"{work}ing results will be saved in {new_path}")
    return new_path